/* ===========================================================================

Application Code for Billboard tutorial
Author : Byungil Jeong

This source is based on brick shader tutorial written by 3Dlabs Inc.

Copyright (C) 2002-2005  3Dlabs Inc. Ltd.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

    Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

    Redistributions in binary form must reproduce the above
    copyright notice, this list of conditions and the following
    disclaimer in the documentation and/or other materials provided
    with the distribution.

    Neither the name of 3Dlabs Inc. Ltd. nor the names of its
    contributors may be used to endorse or promote products derived
    from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

=========================================================================== */

//
// The source of the shaders itself are in plain text files. Files ending in
// .vert are vertex shaders, and files ending in .frag are fragment shaders.
//

#include "billboard.h"
#include <wand/magick_wand.h>

//
// Handle of the window we're rendering to
//
static GLint window;

GLuint billboardProg;
GLuint bg[5], boardTex;

//
// Movement variables
//
float fXDiff = 0;
float fXInertia = -0.2;
float fScale = 1.0;
int Rotate = 1;

float aspect = 1.0;
int textureID = 0;
const float cameraInitPos[3] = {0.0, -1.5, 7.0};
float cameraPos[3];

#define SCALE_INCREMENT		0.02f
#define CAMPOS_DEV			0.1f
#define TIMER_FREQUENCY_MILLIS  20

/*public*/
int printOglError(char *file, int line)
{
    //
    // Returns 1 if an OpenGL error occurred, 0 otherwise.
    //
    GLenum glErr;
    int    retCode = 0;

    glErr = glGetError();
    while (glErr != GL_NO_ERROR)
    {
        printf("glError in file %s @ line %d: %s\n", file, line, gluErrorString(glErr));
        retCode = 1;
        glErr = glGetError();
    }
    return retCode;
}


/***************************************************************************/
/* Parse GL_VERSION and return the major and minor numbers in the supplied
 * integers.
 * If it fails for any reason, major and minor will be set to 0.
 * Assumes a valid OpenGL context.
*/

void getGlVersion( int *major, int *minor )
{
    const char* verstr = (const char*)glGetString( GL_VERSION );
    if( (verstr == NULL) || (sscanf( verstr, "%d.%d", major, minor ) != 2) )
    {
        *major = *minor = 0;
        fprintf( stderr, "Invalid GL_VERSION format!!!\n" );
    }
}

/******************************************************************************

 GLUT glue

******************************************************************************/

static
void display(void)
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	// set perspective projection
	glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
	gluPerspective(60.0, aspect, 1, 1000);
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();

	// using OpenGL - disable shaders
	glUseProgram(0);
	glBindTexture(GL_TEXTURE_2D, bg[textureID]);

	glPushMatrix();
	// set camera for background
	gluLookAt(0.0,0.0,0.0, 0.0,0.0,-1.0, 0.0f,1.0f,0.0f);
	glTranslatef(0, 0, -100);
	glScalef(60.0*aspect, 60.0, 0.0);

	// draw background
	glBegin( GL_POLYGON );
		glTexCoord2f( 0.0f, 0.0f );
		glVertex3f( -1.0f, -1.0f, 0.0f);

		glTexCoord2f( 1.0f, 0.0f );
		glVertex3f(  1.0f, -1.0f, 0.0f);

		glTexCoord2f( 1.0f, 1.0f );        
		glVertex3f(  1.0f,  1.0f, 0.0f);

		glTexCoord2f( 0.0f, 1.0f );     
		glVertex3f( -1.0f,  1.0f, 0.0f);
	glEnd();    
	glPopMatrix();

	//set camera for objects
   gluLookAt(cameraPos[0], cameraPos[1], cameraPos[2], 0.0, 0.0, 2.0, 0.0f, 1.0f, 0.0f);
		 
	glRotatef(fXDiff, 0,1,0);

	glScalef(fScale, fScale, fScale);
	
	// draw billboard
	glUseProgram(billboardProg);
	glBindTexture(GL_TEXTURE_2D, boardTex);
	glUniform1i(getUniLoc(billboardProg, "drawPost"), 0);

	// draw a cube with appropriate texture coordinates and surface normals
	drawCube();
	
	// draw a post
	glUniform1i(getUniLoc(billboardProg, "drawPost"), 1);
	glTranslatef(0.0f, -5.0f, 0.0f);
	glScalef(0.4f, 10.0f, 0.4f);
	glutSolidCube(1.0f);
	
	glFlush();
	glutSwapBuffers();
}


static
void play(void)
{
    glutPostRedisplay();
}


static
void key(unsigned char keyPressed, int x, int y)
{
    switch(keyPressed)
    {
        case 'q':
        case 27:
            exit(0);
            break;
        case 't':
            //textureID = (textureID + 1) % 2;
            break;
        case ' ':
            Rotate = !Rotate;
            break;
        case '+':
            fScale += SCALE_INCREMENT;
            break;
        case '-':
            fScale -= SCALE_INCREMENT;
            break;
        default:
            fprintf(stderr, "\nKeyboard commands:\n\n"
            "q, <esc> - Quit\n"
            "t - Toggle background image\n"
            "? - Help\n"
            "<home>     - reset zoom and rotation\n"
            "<space>    - stop rotation\n"
				"arrow keys - move camera left/right/up/down\n"
            "<+>, <->   - zoom in/out\n");
            break;
    }
}

static
void timer(int value)
{
    /* Increment wrt inertia */
    if (Rotate)
    {
        fXDiff = (fXDiff + fXInertia);
    }

    /* Callback */
    glutTimerFunc(TIMER_FREQUENCY_MILLIS , timer, 0);
}

static
void reshape(int wid, int ht)
{
    aspect = (float) wid / (float) ht;

    glViewport(0, 0, wid, ht);
}


static
void special(int key, int x, int y)
{
	switch (key) {
		case GLUT_KEY_HOME:
   		fXDiff    = 0;
   		fXInertia = -0.2;
   		fScale    = 1.0;
			cameraPos[0] = cameraInitPos[0];
			cameraPos[1] = cameraInitPos[1];
			cameraPos[2] = cameraInitPos[2];    
		break;
		case GLUT_KEY_LEFT:
			cameraPos[0] -= CAMPOS_DEV;
		break;
		case GLUT_KEY_RIGHT:
			cameraPos[0] += CAMPOS_DEV;
		break;
		case GLUT_KEY_UP:
			cameraPos[1] += CAMPOS_DEV;
		break;
		case GLUT_KEY_DOWN:
			cameraPos[1] -= CAMPOS_DEV;
		break;
	}
}

// read an image using ImageMagick and load it to texture
GLuint loadTexture(char *fileName)
{
	MagickBooleanType status;  
	MagickWand *magick_wand;
	
	magick_wand=NewMagickWand();  
	status = MagickReadImage(magick_wand, fileName);

	if (status == MagickFalse) {
		printf("Fail to read image %s\n", fileName);
		return 1;
	}
		
	int width = MagickGetImageWidth(magick_wand);
	int height = MagickGetImageHeight(magick_wand);

	unsigned char *pixels = (unsigned char*)malloc(width*height*3);

	// convert row order
	int i;
	for (i=0; i<height; i++)
		MagickGetImagePixels(magick_wand, 0, i, width, 1, "RGB", CharPixel, 
			&pixels[(height-i-1)*width*3]);
		
	DestroyMagickWand(magick_wand);
	
	GLuint texID = 0;  
	glGenTextures(1, &texID);
	glBindTexture(GL_TEXTURE_2D, texID);
	
   glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, GL_RGB,
        GL_UNSIGNED_BYTE, (const GLvoid *) pixels);

	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	
	return texID;
}

int installBillboardShaders(const GLchar *billboardVertex,
                        const GLchar *billboardFragment)
{
	 billboardProg = installShaders(billboardVertex, billboardFragment); // Program handle

    if (!billboardProg)
        return 0;

    // Install program object as part of current state

    glUseProgram(billboardProg);

    // Set up initial uniform values
    glUniform3f(getUniLoc(billboardProg, "mortarColor"), 0.85, 0.86, 0.84);
	 glUniform1i(getUniLoc(billboardProg, "colorMap"), 0);
	 glUniform1i(getUniLoc(billboardProg, "drawPost"), 0);

    return 1;
}

/******************************************************************************

 Main

******************************************************************************/

int main( int argc, char **argv )
{
	int gl_major, gl_minor;
	GLchar *VertexShaderSource, *FragmentShaderSource;

	glutInit( &argc, argv );
	glutInitDisplayMode( GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
	glutInitWindowSize(512, 512);
	window = glutCreateWindow( "CS594 - Billboard Tutorial");

	glutIdleFunc(play);
	glutDisplayFunc(display);
	glutKeyboardFunc(key);
	glutReshapeFunc(reshape);
	glutSpecialFunc(special);
	glutTimerFunc(TIMER_FREQUENCY_MILLIS , timer, 0);

	// Initialize the "OpenGL Extension Wrangler" library
	glewInit();

	// Make sure that OpenGL 2.0 is supported by the driver
	getGlVersion(&gl_major, &gl_minor);
	printf("GL_VERSION major=%d minor=%d\n", gl_major, gl_minor);

	if (gl_major < 2) {
		printf("GL_VERSION major=%d minor=%d\n", gl_major, gl_minor);
		printf("Support for OpenGL 2.0 is required for this demo...exiting\n");
		exit(1);
	}

	glDepthFunc(GL_LESS);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_TEXTURE_2D); // enable 2D texture
		     
	key('?', 0, 0);     // display help

	// load shaders
	readShaderSource("billboard", &VertexShaderSource, &FragmentShaderSource);
	if (!installBillboardShaders(VertexShaderSource, FragmentShaderSource))
		return 0;

	boardTex = loadTexture("hummer.jpg");	 
	// load environment texture	
	bg[0] = loadTexture("sky1.jpg");

	cameraPos[0] = cameraInitPos[0];
	cameraPos[1] = cameraInitPos[1];
	cameraPos[2] = cameraInitPos[2];    
	glutMainLoop();
    
	return 0;
}
